#################################################################
# Defines the procedures used to create a graph jpg using gnuplot
# and the graphpoints output from Landslide Tcl API
#
# Example usage:
#
#  set gp1 [ls::retrieve graphPoints $test -tab "Test Summary" \
#            -name "Sessions Pending" ]
#  set gp2 [ls::retrieve graphPoints $test -tab "Test Summary" \
#            -name "Sessions Established" ]
#  ls::setGnuplotExecutable C:/gnuplot/bin/wgnuplot.exe"
#  ls::setGraphTitle "My Graph"
#  ls::setGraphFile C:/mydata/graph.jpg
#  ls::addGnuplotMeasurement $gp1 "Sessions Pending"
#  ls::addGnuplotMeasurement $gp2 "Sessions Established"
#  ls::writeGnuplotFile
#  ls::executeGnuplot
#
#
##################################################################

namespace eval ls {

variable graphTitle "Landslide Graph"
variable gnuplotExecutable ""
variable tmpGnuplotFile "TempGnuplotInput.plt"
variable graphFile ""
variable graphData { }
variable graphLabels { }

proc setGraphTitle { name } {
  variable graphTitle
  set graphTitle $name
}

proc setGnuplotExecutable { pathname } {
  variable gnuplotExecutable
  set gnuplotExecutable $pathname
}

proc setTmpGnuplotFile { pathname } {
  variable tmpGnuplotFile
  set tmpGnuplotFile $pathname
}

proc setGraphFile { pathname } {
  variable graphFile
  set graphFile $pathname
}

#################################################################
# Resets the graph points for the gnuplot
##################################################################
proc resetGraphPoints { } {
  variable graphData
  variable graphLabels
  set graphData { }
  set graphLabels { }
}

#################################################################
# Adds a measurement to the current gnuplot
# 
# measData - List of graph data points:
#    { { g1x1 g1y1 } ... { g1xn g1yn} }
# measLabel - The name of each graph line
##################################################################
proc addGnuplotMeasurement { measData measLabel  } {
  variable graphData
  variable graphLabels 
 
  if { [llength $measData] < 3 } {
    set v [llength $measData] 
    return "Invalid Measurement Data only $v points"
  }
  if { "$measLabel" == "" } {
     return "Measurement label is empty"
  }

  lappend graphData $measData
  lappend graphLabels $measLabel
}

#################################################################
# Executes gnuplot at $ls::gnuplotExecutable passing the temporary
# gnuplot input file at $ls::tmpGnuplotFile
##################################################################
proc executeGnuplot { } {
  variable gnuplotExecutable 
  variable tmpGnuplotFile

  if { "$gnuplotExecutable" == "" } {
      return "Gnuplot executable isn't set, set ls::gnuplotExecutable pathname"
  }
  if { [ file isfile $gnuplotExecutable ] == 0 || [ file executable $gnuplotExecutable ] == 0 } {
      return "Gnuplot executable file, $gnuplotExecutable, isn't executable"
  }
  if { [file isfile "$tmpGnuplotFile" ] == 0 } {
     return "Temp plot file does not exist, set ls::gnuplotExecutable"
  }
      
  puts "Calling gnuplot...."
  exec $gnuplotExecutable $tmpGnuplotFile
}


#################################################################
# Method to convert Landslide Tcl API graphpoints output into
# gnuplot input file.
#
# Requires that a least one measurement has been added with 
# ls::addGnuplotMeasurement
#
##################################################################
proc writeGnuplotFile { } {
  variable tmpGnuplotFile
  variable graphFile
  variable graphTitle
  variable graphData
  variable graphLabels
  
  if { [llength $graphData] == 0 || [llength [lindex $graphData 0]] < 3 } {
    set v [llength $graphData] 
    set p  [llength [lindex $graphData 0]] 
    return "Invalid Arguments $v $p"
  }
  if { [llength $graphData] != [llength $graphLabels] } {
    return "Invalid Arguments, graphData != graphLabels"
  }
  if { "$tmpGnuplotFile" == "" } {
     return "Temporary Gnuplot File pathname is not set, set ls::tmpGnuplotFile pathname"
  }

  # open the filename for writing
  set fileId [open $tmpGnuplotFile "w"]

  #output terminal type jpeg to enable jpg writing
  puts $fileId "set term jpeg"

  #set the output file
  puts $fileId "set output \"$graphFile\""

  #set the title
  if { $graphTitle != "" } {
    puts $fileId "set title \"$graphTitle\""
  }

  #write out the plot data points with lines and labels
  puts -nonewline $fileId "plot "
  set numberOfLines [llength $graphLabels]
  for {set line 0} {$line<$numberOfLines} {incr line} {
    set lineLabel [lindex $graphLabels $line]
    if { $line != 0 } {
      puts -nonewline $fileId ","
    }
    puts -nonewline $fileId "'-' w lines title \"$lineLabel\" "
  }
  puts $fileId ""
    
  set numberOfLines [llength $graphData]
  for {set line 0} {$line<$numberOfLines} {incr line} {
    set lineData [lindex $graphData $line]
    set numberOfPoints [llength $lineData]
    for {set i 0} {$i<$numberOfPoints} {incr i} {
      set point [lindex $lineData $i]
      set x [lindex $point 0]
      set y [lindex $point 1]
      puts $fileId "$x $y"
    }
    puts $fileId "e"
  }
  # close the file, ensuring the data is written out before you continue
  #  with processing.
  close $fileId
}

}
## End namespace
