#!/usr/lib/python

""" Landslide Python to Tcl API Bridge Module 

To properly run this module you must first fully install the Landslide
Tcl API.   For minimal effort we recommend using the Standalone Tcl API
installation instructions.   Once confirmed you can run and login to the
Tcl API, then running this module in Python should be simple.

To run this module, just import the module and/or LsApi/LsApiShell directly,
then pass in the path to your 32-bit JRE in the jre_path keyword to the 
LsApi/LsApiShell constructor.  You will have to either update your PYTHONPATH
or os.chdir do this directory to import the module.  
NOTE: Requires the same version of Java that the rest of Landslide requires

Example usage:

from lsapi import LsApi
username='myuser'
password='mypass'
ip='1.1.1.1'

ls = LsApi(jre_path='/home/myuser/OpenJDK_ValidVersion')
ls.tcl("ls::login " + ip + " " + username + " " + password)

WARNING: this module, as file lsapi.py, will be overwritten each time you upgrade
your Landslide Tcl API.  If you modify this for your own use, be sure to rename
it and remove/change this docstring header block.
"""

__copyright__ = "Copyright 2023, Spirent Communications"

import time
import os
import sys
import subprocess
import fcntl
import logging
import logging.handlers

filepath = os.path.abspath(__file__)
dir_path = os.path.dirname(filepath)

class LsApiShell(subprocess.Popen):
    """ Python access to Tcl API using an extra puts for determining command completion """
    
    POLL_INTERVAL = 0.1
    TIMEOUT_S = 10
    
    def __init__(self, *args, **kwargs):
        """Construct interactive Popen.
        
        Expects the jre_path keyword set to your 32-bit JRE.
        Accepts other POpen constructor keywords as well
        
        """
        
        #JRE_PATH="<MY_PATH>/jdk1.8.0_45/jre"
        JRE_PATH = kwargs.get("jre_path")
        if JRE_PATH is None:
           raise ValueError('jre_path is require keyword, pointed to root of your 32-bit Java JRE')
        
        if 'jre_path' in kwargs:
            del kwargs['jre_path']
        
        # setup the environment
        environ = os.environ.copy()
        environ['JAVA_MAX_MEMORY']='512m'
        environ['LD_LIBRARY_PATH']=dir_path + '/lib:' + JRE_PATH + '/lib/i386:' + JRE_PATH + '/lib/i386/client:' + JRE_PATH + '/lib/i386/native_threads:' + dir_path + '/lib/tcljava1.4.1'
        ldpath = os.environ.get('LD_LIBRARY_PATH', '')
        if ldpath == '':
            environ['LD_LIBRARY_PATH'] = environ['LD_LIBRARY_PATH'] + ':' + ldpath
        environ['CLASSPATH']=dir_path + '/landslidetclclient.jar:' + dir_path + '/coast_client.jar:' + dir_path + '/coast_shared.jar:' + dir_path + '/sp_product.jar'
        environ['TCLLIBPATH']=dir_path + '/lib:' + dir_path + '/lib/tcl8.4'
        environ['PATH']='dir_path +/lib/tcl8.4:'+ os.environ['PATH']
        environ['XP_TCLJAVA_INSTALL_DIR']=dir_path + '/lib/tcljava1.4.1'
        
        # pass through the keywords
        keyword_args = {
            'stdin': subprocess.PIPE,
            'stdout': subprocess.PIPE,
            'stderr': subprocess.STDOUT,
            'env': environ
        }
        keyword_args.update(kwargs)
        
        # create logger
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.INFO)
        self.logger.propagate = False
        
        # create file handler, with a formatter and set level 
        ch = logging.handlers.RotatingFileHandler(dir_path + "/LsApi.log", mode='a', maxBytes=1000000, backupCount=1)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        ch.setFormatter(formatter)
        ch.setLevel(logging.INFO)
        self.logger.addHandler(ch)
        
        # start Tcl
        subprocess.Popen.__init__(self, dir_path + "/bin/tclsh8.4", **keyword_args)
        
        # Make stderr and stdout non-blocking.
        for outfile in (self.stdout, self.stderr):
            if outfile is not None:
                fd = outfile.fileno()
                fl = fcntl.fcntl(fd, fcntl.F_GETFL)
                fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)
        
        # setup Tcl prompt, initialize Landslide Tcl API
        self.logger.info("CONNECTED")
        self.tcl('')
        self.logger.info("Loading Landslide Tcl API...")
        ret = self.tcl('source ' + dir_path + '/landslidetclapi.tcl')
        if "Debug Logs will be" in ret:
            self.logger.info("Loaded")
        else:
            self.logger.warn("Warning: " + ret)
            print("There may have been problem loading Tcl API, check LsApi.log")
        ret = self.tcl("ls::get ApiOptions")
        if "JREMemoryUsage" in ret:
            self.logger.info("ApiOptions confirmed")
        else:
            self.logger.warn("Warning: " + ret)
            print("There may have been problem loading Tcl API, check LsApi.log")
        
    def tcl(self, text, timeout_s = TIMEOUT_S):
        """Communicate with the child process without closing stdin."""
        self.stdin.write('set ret ['+ text + "]\n")
        prompt = "\n\nLsApiDone##>>\n"
        self.stdin.write("puts $ret\n")
        self.stdin.write("puts \"\\n\\nLsApiDone##>>\"\n")
        self.stdin.flush()
        self.logger.info("SENT:" + text)
        str_buffer = ''
        i = 0
        num_loops = timeout_s / LsApiShell.POLL_INTERVAL
        
        self.logger.info("Waiting for response to:" + text)
        while not str_buffer.endswith(prompt):
            try:            
                str_buffer += self.stdout.read()        
                self.logger.info("Read some data")
            except IOError:
                time.sleep(LsApiShell.POLL_INTERVAL)
                i += 1
                self.logger.debug("IO Error")
                if i > num_loops:
                    if prompt in str_buffer:
                        offset = str_buffer.find(prompt)
                        self.logger.warn("RECEIVED-ERR, offset=" + str(offset) + ": " + str_buffer)
                        return str_buffer[:offset-1]
                    self.logger.warn("RECEIVED-ERR:" + str_buffer)
                    return "TIMEOUT, partial response: " + str_buffer
        offset = str_buffer.find(prompt)
        self.logger.info("RECEIVED:" + str_buffer[:offset-1])
        return str_buffer[:offset-1]


class LsApi(subprocess.Popen):
    """ Python access to Tcl API using a Tcl prompt for determining command completion """
    
    POLL_INTERVAL = 0.1
    TIMEOUT_S = 10
    
    def __init__(self, *args, **kwargs):
        """Construct interactive Popen.
        
        Expects the jre_path keyword set to your 32-bit JRE.
        Accepts other POpen constructor keywords as well
        
        """
        
        #JRE_PATH="<MY_PATH>/jdk1.8.0_45/jre"
        JRE_PATH = kwargs.get("jre_path")
        if JRE_PATH is None:
           raise ValueError('jre_path is require keyword, pointed to root of your 32-bit Java 7+ JRE')
        
        if 'jre_path' in kwargs:
            del kwargs['jre_path']
        
        # setup the environment
        environ = os.environ.copy()
        environ['JAVA_MAX_MEMORY']='512m'
        environ['LD_LIBRARY_PATH']=dir_path + '/lib:' + JRE_PATH + '/lib/i386:' + JRE_PATH + '/lib/i386/client:' + JRE_PATH + '/lib/i386/native_threads:' + dir_path + '/lib/tcljava1.4.1'
        ldpath = os.environ.get('LD_LIBRARY_PATH', '')
        if ldpath == '':
            environ['LD_LIBRARY_PATH'] = environ['LD_LIBRARY_PATH'] + ':' + ldpath
        environ['CLASSPATH']=dir_path + '/landslidetclclient.jar:' + dir_path + '/coast_client.jar:' + dir_path + '/coast_shared.jar:' + dir_path + '/sp_product.jar'
        environ['TCLLIBPATH']=dir_path + '/lib:' + dir_path + '/lib/tcl8.4'
        environ['PATH']='dir_path +/lib/tcl8.4:'+ os.environ['PATH']
        environ['XP_TCLJAVA_INSTALL_DIR']=dir_path + '/lib/tcljava1.4.1'
        
        # pass through the keywords
        keyword_args = {
            'stdin': subprocess.PIPE,
            'stdout': subprocess.PIPE,
            'stderr': subprocess.STDOUT,
            'env': environ
        }
        keyword_args.update(kwargs)
        
        # create logger
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.INFO)
        self.logger.propagate = False
        
        # create file handler, with a formatter and set level to info
        ch = logging.handlers.RotatingFileHandler(dir_path + "/LsApi.log", mode='a', maxBytes=1000000, backupCount=1)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        ch.setFormatter(formatter)
        ch.setLevel(logging.INFO)
        self.logger.addHandler(ch)
        
        # start Tcl
        subprocess.Popen.__init__(self, dir_path + "/bin/tclsh8.4", **keyword_args)
        
        # Make stderr and stdout non-blocking.
        for outfile in (self.stdout, self.stderr):
            if outfile is not None:
                fd = outfile.fileno()
                fl = fcntl.fcntl(fd, fcntl.F_GETFL)
                fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)
        
        # setup Tcl prompt, initialize Landslide Tcl API
        self.logger.info("CONNECTED")
        self.prompt = "\n\nLsApiDone##>>\n"
        self.stdin.write('proc showPrompt { } { puts "\\n\\nLsApiDone##>>" }\n')
        self.stdin.flush()
        self.stdin.write('set tcl_prompt1 showPrompt\n')
        self.stdin.flush()
        self.stdin.write('set tcl_interactive 1\n')
        self.stdin.flush()
        self.tcl('')
        self.logger.info("Loading Landslide Tcl API...")
        ret = self.tcl('source ' + dir_path + '/landslidetclapi.tcl')
        if "Debug Logs will be" in ret:
            self.logger.info("Loaded")
        else:
            self.logger.warn("Warning: " + ret)
            print("There may have been problem loading Tcl API, check LsApi.log")
        ret = self.tcl("ls::get ApiOptions")
        if "JREMemoryUsage" in ret:
            self.logger.info("ApiOptions confirmed")
        else:
            self.logger.warn("Warning: " + ret)
            print("There may have been problem loading Tcl API, check LsApi.log")
        
    def tcl(self, text, timeout_s = TIMEOUT_S):
        """Communicate with the child process without closing stdin."""
        self.stdin.write(text + "\n")
        self.stdin.flush()
        self.logger.info("SENT:" + text)
        str_buffer = ''
        i = 0
        num_loops = timeout_s / LsApi.POLL_INTERVAL
        
        self.logger.info("Waiting for response to:" + text)
        while not str_buffer.endswith(self.prompt):
            try:            
                str_buffer += self.stdout.read()        
                self.logger.info("Read some data")
            except IOError:
                time.sleep(LsApi.POLL_INTERVAL)
                i += 1
                self.logger.debug("IO Error")
                if i > num_loops:
                    if self.prompt in str_buffer:
                        offset = str_buffer.find(prompt)
                        self.logger.warn("RECEIVED-ERR, offset=" + str(offset) + ": " + str_buffer)
                        return str_buffer[:offset-1]
                    self.logger.warn("RECEIVED-ERR:" + str_buffer)
                    return "TIMEOUT, partial response: " + str_buffer
        offset = str_buffer.find(self.prompt)
        self.logger.info("RECEIVED:" + str_buffer[:offset-1])
        return str_buffer[:offset-1]
